import {
  MODE_RANGE, MODE_SINGLE,
  VIEW_MODE_DOUBLE,
  VIEW_MODE_SINGLE, VIEW_MODE_CUSTOM
} from './utils/constants';
import { getDateSelectStrategy } from './utils/get-date-select-strategy';
import {
  shouldPreventMonthChange,
  generateCalendarViewData,
  resetDate,
  getMonthsList
} from './utils/renderless-calendar.service';
import CalendarDate from './classes/CalendarDate';
import {
  getMonthDateSafely,
  isEqual,
  isMarked,
  isNextDate,
  isPrevDate,
  isRestricted
} from './utils/renderless-date.service';
import { viewModeInitializer, getViewNavigator } from './utils/view-mode.service';
import { getCalendarStringsForLocale } from './utils/locale-strings';
/**
 * @typedef {{
 *  months:{short: string, full: string}[],
 *  days:{short: string, full: string}[]
 * }} LocaleStrings
 */

export default {
  name: 'RenderlessCalendar',

  props: {
    startDate: {
      type: [Date, String],
      default: () => new Date()
    },
    viewMode: {
      type: String,
      validator(v) {
        return [VIEW_MODE_SINGLE, VIEW_MODE_DOUBLE, VIEW_MODE_CUSTOM].indexOf(v) !== -1;
      },
      default: VIEW_MODE_SINGLE
    },
    customNumberOfMonths: {
      type: Number,
      default: 0
    },
    mode: {
      type: String,
      validator(value) {
        return [MODE_SINGLE, MODE_RANGE].indexOf(value) !== -1;
      },
      default: MODE_SINGLE
    },
    locale: {
      type: [String, Object],
      default: ''
    },
    minDate: {
      type: String,
      default: null
    },
    maxDate: {
      type: String,
      default: null
    },
    preventOutOfRange: {
      default: true,
      type: Boolean
    },
    captureThirdDate: {
      type: Boolean,
      default: false
    },
    dateSelectStrategy: {
      type: Function,
      default: null
    },
    defaultSelectedDates: {
      type: Array,
      default: () => []
    },
    captureHover: {
      type: Boolean,
      default: true
    },
    disabledDates: {
      type: Array,
      default: () => []
    },
    markedDates: {
      type: Array,
      default: () => []
    },
    firstDayOfWeek: {
      type: Number,
      default: 1
    },
    omitOtherMonthDays: {
      type: Boolean,
      default: false
    }
  },

  data() {
    return {
      calendar: null,
      selectedDates: [],
      currentMonth: null,
      currentYear: null,
      currentHoveredDate: null,
      viewState: []
    };
  },

  created() {
    this.selectedDates = this.defaultSelectedDates.map(CalendarDate.fromString);
    const startDate = typeof this.startDate === 'string' ? new Date(this.startDate) : this.startDate;

    this.currentMonth = startDate.getMonth();
    this.currentYear = startDate.getFullYear();

    this.viewState = viewModeInitializer(this.viewMode, this.customNumberOfMonths, startDate);
    this.viewNavigator = getViewNavigator(this.viewMode);
    this.calendar = Object.freeze(generateCalendarViewData({
      numberOfMonths: this.customNumberOfMonths,
      month: this.currentMonth,
      viewMode: this.viewMode,
      year: this.currentYear,
      firstDayOfWeek: this.firstDayOfWeek,
      omitOtherMonthDays: this.omitOtherMonthDays
    }));
  },

  computed: {
    /**
     * @return {LocaleStrings}
     */
    localeStrings() {
      return typeof this.locale === 'string'
        ? getCalendarStringsForLocale(this.locale, this.firstDayOfWeek)
        : this.locale;
    },
    weekDayNames() {
      return this.localeStrings.days;
    },
    monthNames() {
      return this.localeStrings.months
        .map(({ short, full }, id) => ({ short, full, id }));
    },
    monthsList() {
      return getMonthsList({
        currentMonth: this.currentMonth,
        currentYear: this.currentYear,
        monthNames: this.monthNames,
        viewState: this.viewState
      });
    },
    canGoToPrevMonth() {
      const [{ month, year }] = this.viewState;

      return !shouldPreventMonthChange({
        minDate: this.minDate,
        maxDate: this.maxDate,
        currentMonth: month,
        currentYear: year,
        step: -1
      });
    },
    canGoToNextMonth() {
      const { month, year } = this.viewState[this.viewState.length - 1];

      return !shouldPreventMonthChange({
        minDate: this.minDate,
        maxDate: this.maxDate,
        currentMonth: month,
        currentYear: year,
        step: 1
      });
    }
  },

  methods: {
    onDateSelect(date) {
      this.setDates(this.dateChangeHangler(this.selectedDates, date));
    },
    onDateMouseOver(date) {
      this.currentHoveredDate = date;
    },
    onDateMouseOut() {
      this.currentHoveredDate = null;
    },
    setDates(dates) {
      this.selectedDates = dates;
      this.$emit('onDateChange', dates);
    },
    resetDates(index) {
      this.setDates(resetDate(this.selectedDates, index));
    },
    prevPage(jumpToMonthIndex) {
      const { viewState, viewMode, customNumberOfMonths } = this;
      const [date] = viewState;
      let { year, month } = date;
      let newViewState = [];

      if (jumpToMonthIndex) {
        month = jumpToMonthIndex + 1;
      }

      if (this.shouldPreventMonthChange(year, month, -1) || month < 0) {
        return;
      }

      const prevMonthDate = getMonthDateSafely(year, month - 1);

      const calendar = generateCalendarViewData({
        numberOfMonths: customNumberOfMonths,
        year: prevMonthDate.getFullYear(),
        month: prevMonthDate.getMonth(),
        viewMode,
        firstDayOfWeek: this.firstDayOfWeek,
        omitOtherMonthDays: this.omitOtherMonthDays
      });

      if (viewMode === VIEW_MODE_SINGLE) {
        newViewState = [{
          year: prevMonthDate.getFullYear(),
          month: prevMonthDate.getMonth()
        }];
      } else if (viewMode === VIEW_MODE_DOUBLE) {
        const first = viewState[0];

        newViewState = [{
          year: prevMonthDate.getFullYear(),
          month: prevMonthDate.getMonth()
        }, first];
      }

      this.calendar = calendar;
      this.viewState = newViewState;
    },
    nextPage(jumpToMonthIndex) {
      const { viewMode, viewState, customNumberOfMonths } = this;
      let newViewState = [];
      let { year, month } = viewMode === VIEW_MODE_SINGLE
        ? viewState[0]
        : viewState[1];

      if (jumpToMonthIndex) {
        month = jumpToMonthIndex - 1;
      }
        
      if (this.shouldPreventMonthChange(year, month, 1) || month < 0) {
        return;
      }

      if (viewMode === VIEW_MODE_DOUBLE) {
        const first = viewState[0];
        year = first.year;
        month = first.month;
      }

      const nextMonthDate = getMonthDateSafely(year, month + 1);

      const calendar = generateCalendarViewData({
        numberOfMonths: customNumberOfMonths,
        year: nextMonthDate.getFullYear(),
        month: nextMonthDate.getMonth(),
        viewMode,
        firstDayOfWeek: this.firstDayOfWeek,
        omitOtherMonthDays: this.omitOtherMonthDays
      });

      if (viewMode === VIEW_MODE_SINGLE) {
        newViewState = [{
          year: nextMonthDate.getFullYear(),
          month: nextMonthDate.getMonth()
        }];
      } else if (viewMode === VIEW_MODE_DOUBLE) {
        const second = viewState[1];
        const nextMonthDate = getMonthDateSafely(year, month + 2);

        newViewState = [second, {
          year: nextMonthDate.getFullYear(),
          month: nextMonthDate.getMonth()
        }];
      }

      this.calendar = calendar;
      this.viewState = newViewState;
    },
    shouldPreventMonthChange(year, month, step) {
      return this.preventOutOfRange && shouldPreventMonthChange({
        currentMonth: month,
        currentYear: year,
        minDate: this.minDate,
        maxDate: this.maxDate,
        step
      });
    },
    setMonth(monthListItem) {
      const { month, year } = monthListItem;
      const { viewMode, customNumberOfMonths } = this;
      
      let newViewState = viewModeInitializer(this.viewMode, this.customNumberOfMonths, new Date(year, month + 1));

      this.calendar = generateCalendarViewData({
        numberOfMonths: customNumberOfMonths,
        viewMode,
        month,
        year,
        firstDayOfWeek: this.firstDayOfWeek,
        omitOtherMonthDays: this.omitOtherMonthDays
      });
      this.viewState = newViewState;
    },
    isBetween(date) {
      return this.captureHover && date.isBetween({
        currentHoveredDate: this.currentHoveredDate,
        captureThirdDate: this.captureThirdDate,
        selectedDates: this.selectedDates
      });
    },
    isSelected(date) {
      const isSelectingRange = this.currentHoveredDate && this.mode === MODE_RANGE && this.selectedDates.length === 1;
      return date.isSelected(this.selectedDates)
        || (isSelectingRange && date.formatted === this.currentHoveredDate.formatted);
    },
    isDisabled(date) {
      return isRestricted(date.formatted, {
        disabledDates: this.disabledDates,
        maxDate: this.maxDate,
        minDate: this.minDate
      });
    },
    isOneDayAfter(date) {
      return this.selectedDates
        .map(_ => _.formatted)
        .some(formatted => isNextDate(date.formatted, formatted));
    },
    isOneDayBefore(date) {
      return this.selectedDates
        .map(_ => _.formatted)
        .some(formatted => isPrevDate(date.formatted, formatted));
    },
    isOneDayBeforeFirst(date) {
      const first = this.selectedDates[0];
      return first && isPrevDate(date.formatted, first.formatted);
    },
    isOneDayAfterFirst(date) {
      const first = this.selectedDates[0];
      return first && isNextDate(date.formatted, first.formatted);
    },
    isOneDayBeforeLast(date) {
      const last = this.selectedDates[1];
      return last && isPrevDate(date.formatted, last.formatted);
    },
    isOneDayAfterLast(date) {
      const last = this.selectedDates[1];
      return last && isNextDate(date.formatted, last.formatted);
    },
    isFirst(date) {
      const first = this.selectedDates[0];
      return first && isEqual(date.formatted, first.formatted);
    },
    isLast(date) {
      const last = this.selectedDates[1];
      return last && isEqual(date.formatted, last.formatted);
    },
    isMarked(date) {
      return isMarked(date.formatted, this.markedDates);
    },
    getModifiers(date) {
      return {
        isBetween: this.isBetween(date),
        isSelected: this.isSelected(date),
        isDisabled: this.isDisabled(date),
        isOneDayAfter: this.isOneDayAfter(date),
        isOneDayBefore: this.isOneDayBefore(date),
        isOneDayBeforeFirst: this.isOneDayBeforeFirst(date),
        isOneDayAfterFirst: this.isOneDayAfterFirst(date),
        isOneDayBeforeLast: this.isOneDayBeforeLast(date),
        isOneDayAfterLast: this.isOneDayAfterLast(date),
        isFirst: this.isFirst(date),
        isLast: this.isLast(date),
        isMarked: this.isMarked(date)
      };
    }
  },

  watch: {
    mode: {
      handler(mode) {
        this.dateChangeHangler = this.dateSelectStrategy || getDateSelectStrategy(mode);
      },
      immediate: true
    },
    firstDayOfWeek(newFirstDoW) {
      const [{ month, year }] = this.viewState;
      this.calendar = Object.freeze(generateCalendarViewData({
        numberOfMonths: this.customNumberOfMonths,
        viewMode: this.viewMode,
        firstDayOfWeek: newFirstDoW,
        omitOtherMonthDays: this.omitOtherMonthDays,
        month,
        year
      }));
    },
    defaultSelectedDates(dates) {
      this.setDates(dates.map(CalendarDate.fromString));
    }
  },

  render() {
    return this.$scopedSlots.default({
      weekDayNames: this.weekDayNames,
      monthNames: this.monthNames,
      monthsList: this.monthsList,

      calendar: this.calendar,
      selectedDates: this.selectedDates,
      currentMonth: this.currentMonth,
      currentYear: this.currentYear,

      canGoToPrevMonth: this.canGoToPrevMonth,
      canGoToNextMonth: this.canGoToNextMonth,

      prevPage: this.prevPage,
      nextPage: this.nextPage,
      resetDates: this.resetDates,
      setMonth: this.setMonth,

      onDateMouseOut: this.onDateMouseOut,
      onDateMouseOver: this.onDateMouseOver,
      onDateSelect: this.onDateSelect,

      getModifiers: this.getModifiers
    });
  }
};
